/*******************************************************************************
* Licensed Materials - Property of HCL
* (c) Copyright HCL Technologies Ltd. 2017.  All Rights Reserved.
*
* Note to U.S. Government Users Restricted Rights:
* Use, duplication or disclosure restricted by GSA ADP Schedule
* Contract with IBM Corp.
*******************************************************************************/

/**
 * Perform steps necessary before the Design Room Live can be installed:
 * - download Mongo DB
 * - download PM2
 *
 * @author Mattias Mohlin
 */

'use strict';

let request = require('request');
let conf = require('nconf');
require('nconf-strip-json-comments')(conf);
let fs = require('fs');
let progress = require('request-progress');
let path = require('path');
let term = require('terminal-kit').terminal;

conf.argv().env({lowerCase: true});
let configFile = conf.get('dr_config_file');
if (configFile) {
    if (!fs.existsSync(configFile)) {
        console.log('Failed to find configuration file specified by dr_config_file: ' + configFile);
        process.exit(1);
    }
    conf.file('custom_config', configFile);
}
conf.file('default_config', {
    file: './dr-config.json'
});

term.on('key', function (name, matches, data) {
    // Detect CTRL-C and exit 'manually'
    if (name === 'CTRL_C') { process.exit(); }
});

// Downloads from a URL into a local file
function download(downloadName, url, file) {
    function doDownload(downloadName, url, file, callback) {
        console.log('Downloading ' + downloadName + ' from ' + url);

        let progressBar = term.progressBar({
            width: 80,
            title: 'Downloaded:',
            percent: true,
            inline: true,
            syncMode: true
        });
        term.up(1);

        progress(request(url), {})
            .on('progress', function (state) {
                progressBar.update(state.percent);
            })
            .on('error', function (err) {
                console.log('Failed to download ' + downloadName);
                console.log('Reason: ' + err);
                process.exit(1);
            })
            .on('end', function () {
                term.down(1);
                console.log('Successfully downloaded ' + downloadName);
                callback();
            })
            .pipe(fs.createWriteStream(file));
    }

    return new Promise((resolve, reject) => {
        if (fs.existsSync(file)) {
            term(file + ' already exists. Download it anyway? [y|N]\n');
            term.yesOrNo({ yes: ['y', 'Y'], no: ['n', 'N', 'ENTER'] }, function (error, result) {
                if (!result) {
                    console.log('Skipping download of ' + downloadName);
                    resolve(false);
                    return;
                }
                doDownload(downloadName, url, file, () => {
                    resolve(true);
                });
            });
        }
        else {
            doDownload(downloadName, url, file, () => {
                resolve(true);
            });
        }
    });

}

// Downloads an npm package 'pkg' and boxes it into a local npmbox file
function npmbox(pkg, file, callback) {
    function doDownload(pkg, file, callback) {
        console.log('Downloading ' + pkg + '...');

        let npmboxLocation = path.join(__dirname, 'node_modules', 'npmbox', 'bin', 'npmbox');
        let spawn = require('child_process').spawn;

        let childProcess = spawn('node', [npmboxLocation, '-t', file, pkg]);    
    
        childProcess.on('error', function (err) {
            console.log('Failed to download ' + pkg + '!');
            process.exit(1);
        });

        childProcess.stdout.on('data', function (data) {
            console.log('' + data);
        });

        childProcess.stderr.on('data', function (data) {
            console.log('' + data);
        });

        childProcess.on('close', function (code) {
            if (code != 0) {
                console.log('Process of boxing ' + pkg + ' terminated unexpectedly!');
                process.exit(1);
            }
            console.log('Successfully downloaded ' + pkg + '!');
            callback();
        });    
    }

    let fileName = file + '.npmbox';
    if (fs.existsSync(fileName)) {
        term(file + ' already exists. Download it anyway? [y|N]\n');
        term.yesOrNo({ yes: ['y', 'Y'], no: ['n', 'N', 'ENTER'] }, function (error, result) {
            if (!result) {
                console.log('Skipping download of ' + pkg);
                callback();
                return;
            }
            fs.unlink(fileName); // Need to delete the existing file first
            doDownload(pkg, file, callback);
        });
    }
    else {
        doDownload(pkg, file, callback);
    }
}

// Download PM2
function downloadPM2() {    
    return new Promise((resolve, reject) => {
        npmbox('pm2', 'downloaded/pm2', () => {
            resolve();
        });  
    });     
}

if (process.platform == 'win32') {
    // Download Mongo on Windows
    var mongoDownload = download('Mongo (Windows)', conf.get('mongo_win_download_url'), 'downloaded/mongo-win.msi');
}
else if (process.platform == 'linux') {  
    // Do not download Mongo on Linux
    var mongoDownload = Promise.resolve(false);    
}
else {
    console.log('Unsupported platform. Design Room Live can only be installed on Windows and Linux systems.');
    process.exit(1);
}

mongoDownload.then((downloaded) => {    
    return downloadPM2();
})
.then( () => {
    if (process.platform == 'linux') {
        term('NOTE: Mongo DB needs to be manually installed on Linux systems. Follow instructions for your Linux distribution at ');
        term.inverse('https://docs.mongodb.com/manual/administration/install-on-linux/#recommended-linux-install'); term('\n');
    }
    term('All done! If you want to install DesignRoomLive on this machine you may run\n');
    term.inverse('node install-dr.js'); term(' now. If you want to install on another machine, copy the\n');
    term('Extracted folder i.e. DesignRoomLIVE_V1.0-BETA_* to that machine and then run '); term.inverse('node install-dr.js'); term(' there.\n');
    process.exit(0);
});





